#include "isr.h"
#include "idt.h"
#include "i386.h"
#include "gdt.h"
#include "klib.h"
#include "loader.h"
#include "consola.h"

idt_entry idt[256];
idt_descriptor IDT_DESC = {sizeof(idt)-1, (unsigned int)&idt};

const char *exception_messages[] = {
	"Divide Error",
	"Reserved",
	"Non-maskeable Interrupt",
	"Breakpoint",
	"Overflow",
	"Index out of range",
	"Invalid Opcode",
	"Buy a coprocessor, support Intel(R)",
	"Double Fault",
	"Coprocessor Segment Overrun",
	"Invalid TSS",
	"Segment not present",
	"Stack segment fault",
	"GPF (all your data are belong to us)",
	"Page Fault",
	"Reserved",
	"FPU Error",
	"Alignment Check",
	"Machine Check",
	"SIMD Floating Point"
};

LS_INLINE void instalar_isr(int num, void (*func)(), unsigned short attr) {
	idt[num].offset_0_15 = (unsigned short) ((unsigned int)(func) & (unsigned int) 0xFFFF); 
	idt[num].segsel = GDT_SEL_KRNCODE;
	idt[num].attr = attr;
	idt[num].offset_16_31 = (unsigned short) ((unsigned int)(func) >> 16 & (unsigned int) 0xFFFF);
}

void inicializar_idt(void) {
	instalar_isr(0, _isr0, 0x8E00);
	instalar_isr(1, _isr1, 0x8E00);
	instalar_isr(2, _isr2, 0x8E00);
	instalar_isr(3, _isr3, 0x8E00);
	instalar_isr(4, _isr4, 0x8E00);
	instalar_isr(5, _isr5, 0x8E00);
	instalar_isr(6, _isr6, 0x8E00);
	instalar_isr(7, _isr7, 0x8E00);
	instalar_isr(8, _isr8, 0x8E00);
	instalar_isr(9, _isr9, 0x8E00);
	instalar_isr(10, _isr10, 0x8E00);
	instalar_isr(11, _isr11, 0x8E00);
	instalar_isr(12, _isr12, 0x8E00);
	instalar_isr(13, _isr13, 0x8E00);
	instalar_isr(14, _isr14, 0x8E00);
	instalar_isr(15, _isr15, 0x8E00);
	instalar_isr(16, _isr16, 0x8E00);
	instalar_isr(17, _isr17, 0x8E00);
	instalar_isr(18, _isr18, 0x8E00);
	instalar_isr(19, _isr19, 0x8E00);
	instalar_isr(32, _isrclock, 0x8E00);
	instalar_isr(33, _isrteclado, 0x8E00);
	instalar_isr(80, _isrsyscall, 0xEE00);
}

/* This defines what the stack looks like after an ISR was running */
struct regs {
	unsigned int edi, esi, ebp, esp, ebx, edx, ecx, eax;  /* pushed by 'pushad' */
	unsigned int int_no, err_code;    /* our 'push byte #' and ecodes do this */
	unsigned int eip, cs, eflags, useresp, ss;   /* pushed by the processor automatically */ 
};

void interrupt_handler(struct regs *r)
{
	kassert(r->int_no < 32);

	const char *msg = (r->int_no < ARRAY_SIZEOF(exception_messages)) ? exception_messages[r->int_no] : "Reserved";

	// TODO: hacer una rutina que (usando snprintf()) sirva para cuando no inicializamos las consolas

	cons_printf(getcons(), COLOR(FG_LIGHT(COL_MARRON), COL_NEGRO),
	    "\nExcepcion %d: %s (codigo de error %d)\n", r->int_no, msg, r->err_code);
	cons_printf(getcons(), COLOR(FG_LIGHT(COL_VERDE), COL_ROJO),
	    "EAX = %x, EBX = %x, ECX = %x, EDX = %x,\nEBP = %x, ESP = %x, "
	    "EDI = %x, ESI = %x,\nEIP = %x, EFLAGS = %x\n", r->eax, r->ebx, r->ecx, r->edx, r->ebp, r->esp,
	    r->edi, r->esi, r->eip, r->eflags);
	
	if (r->cs == GDT_SEL_USRCODE) {
		// Si era una tarea en modo usuario, imprimimos un mensaje y la cerramos
		upanic(NULL);
	} else {
		// Si no, detenemos todo el kernel
		kpanic(NULL);
	}
}
