#include "mmu.h"
#include "loader.h"
#include "i386.h"

int map_kernel[CANT_PAGINAS_KERNEL / 32];
bitmap_t bitmap_kernel = { INICIO_PAGINAS_KERNEL, TAMANO_PAGINA, CANT_PAGINAS_KERNEL, CANT_PAGINAS_KERNEL, map_kernel };

int map_usuario[CANT_PAGINAS_USUARIO / 32];
bitmap_t bitmap_usuario = { INICIO_PAGINAS_USUARIO, TAMANO_PAGINA, CANT_PAGINAS_USUARIO, CANT_PAGINAS_USUARIO, map_usuario };

// XXX: este mapa de memoria solo se usa en el kernel, es el unico que usa paginas de 4MB y nunca se modifica
void mmu_inicializar(void) {
	// Inicializar manejo de memoria
	for (int i = 0; i < ARRAY_SIZEOF(map_kernel); i++)
		map_kernel[i] = 0;
	for (int i = 0; i < ARRAY_SIZEOF(map_usuario); i++)
		map_usuario[i] = 0;

	// Crear el directorio inicial
	dirPagsKernel = obtener_pagina_kernel();
	pdir_entry* dirp = dirPagsKernel;

	// Hacer identity mapping de toda la memoria
	for (int i = 0; i < (MAX_MEM / TAMANO_PAGINA_4MB); i++) {
		dirp[i].base = i * 1024;
		dirp[i].size = 1;
		dirp[i].p = 1;
		dirp[i].rw = 1;
		dirp[i].us = 0;
	}
	// Y marcar el resto como vacia
	for (int i = (MAX_MEM / TAMANO_PAGINA_4MB); i < 1024; i++)
		dirp[i].p = 0;

	// Habilitar paginacion
	lcr3((int)dirPagsKernel);
	lcr0(rcr0() | 0x80000000);
}

void mmu_map(void *dirPags, void *virtual, void *fisica, unsigned attrs) {
	pdir_entry* dirp = dirPags;
	ptbl_entry* tblp = NULL;
	unsigned dir_ind = ((unsigned)virtual >> 22) & 0x3FF;
	unsigned tbl_ind = ((unsigned)virtual >> 12) & 0x3FF;

	if (dirp[dir_ind].p) {
		// El segundo nivel ya existia
		tblp = (void*) (dirp[dir_ind].base << 12);
	} else {
		// Creo una tabla de segundo nivel
		((int*) dirp)[dir_ind] = 0;
		tblp = obtener_pagina_kernel();

		// Inicializar todas las paginas a vacias
		for (unsigned j = 0; j < 1024; j++)
			((int*) tblp)[j] = 0;

		// Llenar la nueva entrada en el directorio
		dirp[dir_ind].base = (unsigned)tblp >> 12;
		dirp[dir_ind].us = 1;
		dirp[dir_ind].rw = 1;
		dirp[dir_ind].p = 1;
	}

	// Agrego la pagina en la tabla (creada o ubicada)
	((int*)tblp)[tbl_ind] = 0;
	tblp[tbl_ind].base = (unsigned)fisica >> 12;
	tblp[tbl_ind].us = SET_TEST(attrs, MMU_KernUser);
	tblp[tbl_ind].rw = SET_TEST(attrs, MMU_ReadWrite);
	tblp[tbl_ind].p = 1;
	
	if (rcr3() == (int)dirPags)
		tlbflush();
}

void mmu_unmap(void *dirPags, void *virtual) {
	pdir_entry* dirp = dirPags;
	ptbl_entry* tblp = NULL;
	unsigned dir_ind = ((unsigned)virtual >> 22) & 0x3FF;
	unsigned tbl_ind = ((unsigned)virtual >> 12) & 0x3FF;

	kassert(dirp[dir_ind].p);
	tblp = (void*) (dirp[dir_ind].base << 12);
	kassert(tblp[tbl_ind].p);
	tblp[tbl_ind].p = 0;
}

void mmu_mapCont(void *dirPags, void *virtual, void *fisica, unsigned attrs, int cantidad) {
	unsigned posVirtual = (unsigned)virtual;
	unsigned posFisica = (unsigned)fisica;

	for (int i = 0; i < cantidad; i++) {
		mmu_map(dirPags, (void*)posVirtual, (void*)posFisica, attrs);
		posVirtual += TAMANO_PAGINA;
		posFisica += TAMANO_PAGINA;
	}
}

void mmu_unmapCont(void *dirPags, void *virtual, int cantidad) {
	unsigned posVirtual = (unsigned)virtual;

	for (int i = 0; i < cantidad; i++) {
		mmu_unmap(dirPags, (void*)posVirtual);
		posVirtual += TAMANO_PAGINA;
	}
}

// Administracion de memoria basica

// TODO: convertir a arbol binario completo en arreglo, con tiempos logaritmicos (perdiendo 2 bits cada 32)

void *bitmap_obtener_pagina(bitmap_t *bitmap) {
	kassert(bitmap->numLibres > 0);
	for (int i = 0; i < bitmap->numTotal / 32; i++) {
		if (bitmap->map[i] == -1)
			continue;
		for (unsigned j = 0, mask = 1; mask != 0; j++, mask <<= 1)
			if ((bitmap->map[i] & mask) == 0) {
				bitmap->map[i] |= mask;
				bitmap->numLibres--;
				return (void*)(bitmap->posInicial + (bitmap->tamElem * (i*32 + j)));
			}
	}
	// Nunca se llega aqui, pero evita warnings
	return 0;
}

void bitmap_liberar_pagina(bitmap_t *bitmap, void *dir) {
	unsigned num = ((int)dir - bitmap->posInicial) / bitmap->tamElem,
		 mask = (unsigned)1 << (num & 31),
		 indice = num / 32;

	// Asegurarse de que todo este en orden, por las dudas
	kassert(dir >= (void*)bitmap->posInicial &&
	    dir < (void*)(bitmap->posInicial + bitmap->numTotal*bitmap->tamElem));
	kassert(((int)dir & (bitmap->tamElem - 1)) == 0);
	kassert((bitmap->map[indice] & mask) != 0);

	bitmap->map[indice] &= ~mask;
	bitmap->numLibres++;
}

void *obtener_pagina_kernel(void) {
	return bitmap_obtener_pagina(&bitmap_kernel);
}

void liberar_pagina_kernel(void *dir) {
	bitmap_liberar_pagina(&bitmap_kernel, dir);
}

void *obtener_pagina_usuario(void) {
	return bitmap_obtener_pagina(&bitmap_usuario);
}

void liberar_pagina_usuario(void *dir) {
	bitmap_liberar_pagina(&bitmap_usuario, dir);
}

// Memoria virtual para procesos

void* crear_dir_usuario(void) {
	pdir_entry* dirp = obtener_pagina_kernel();
	for (int i = 0; i < 1024; i++)
		((int*) dirp)[i] = 0;

	// Parte que usa el kernel, asi andan los system calls y esas cosas
	mmu_mapCont(dirp, (void*)INICIO_KERNEL, (void*)INICIO_KERNEL, PROT_KERNEL_RW, CANT_KERNEL + CANT_PAGINAS_KERNEL);

	return dirp;
}

void destruir_dir_usuario(void *cr3) {
	pdir_entry* dirp = cr3;

	for (int i = 0; i < 1024; i++)
		if (dirp[i].p != 0 && dirp[i].size == 0)
			liberar_pagina_kernel((void*)(dirp[i].base << 12));
	liberar_pagina_kernel(dirp);
}
